@@ -36,11 +36,11 @@ class Photon():
         self.mod_dir = pathlib.Path(__file__).parent.parent
 
         if not self.run_local:
-            from photon import metrics, losses
+            from photon import metrics, losses, utils
             from photon.gamma import Gamma
         else:
             sys.path.append(self.mod_dir)
-            from photon import metrics, losses
+            from photon import metrics, losses, utils
             from photon.gamma import Gamma
 
         self.Networks = Networks
@@ -51,6 +51,7 @@ class Photon():
 
         self.metrics = metrics.Metrics()
         self.losses = losses.Losses()
+        self.utils = utils
 
         self.n_gpus = 0
         self.n_v_gpus = 0
@@ -268,6 +269,8 @@ class Networks():
         if data_cols['t_cols'] is None:
             self.data.t_cols = []
 
+        self.data.f_cols = data_cols['f_cols']
+
         self.data.n_x_cols = len(self.data.x_cols)
         self.data.n_c_cols = len(self.data.c_cols)
         self.data.n_y_cols = len(self.data.y_cols)
@@ -343,6 +346,19 @@ class Networks():
 
         self.data.all_cols = self.data.x_cols + self.data.c_cols + self.data.y_cols + self.data.t_cols
 
+
+        x_ed = self.data.n_x_cols
+        c_st = x_ed
+        c_ed = c_st + self.data.n_c_cols
+        y_st= c_ed
+        y_ed = y_st + self.data.n_y_cols
+        t_st = y_ed
+
+        self.data.slice_configs = {'x_slice': np.s_[..., :x_ed],
+                                   'c_slice': np.s_[..., c_st:c_ed],
+                                   'y_slice': np.s_[..., y_st:y_ed],
+                                   't_slice': np.s_[..., t_st:]}
+
     def setup_x_groups(self, x_cols):
 
         pr_group = []
@@ -443,6 +459,7 @@ class Trees():
                  masking,
                  shuffle,
                  preproc,
+                 outputs_on,
                  seed,
                  name=None,
                  samples_pd=None,
@@ -487,12 +504,11 @@ class Trees():
 
         self.data.shuffle = shuffle
         self.data.preproc = preproc
+        self.data.outputs_on = outputs_on
         self.data.seed = seed
 
         self.data.batch_size = batch_size
 
-        self.data.feature_cols = ['x_cols', 'c_cols']
-
         if samples_pd is None:
 
             self.data.samples_pd = int(23400 / self.data.res)
@@ -606,12 +622,15 @@ class Trees():
         if self.data.seq_agg > 0:
             self.data.seq_depth = int(self.data.seq_len / self.data.seq_agg)
 
-        if 'c_cols' in self.data.feature_cols:
+        if 'c_cols' in self.data.f_cols:
             self.data.input_shape = (self.data.seq_depth, self.data.n_x_cols + self.data.n_c_cols)
 
-        if 'c_cols' not in self.data.feature_cols:
+        if 'c_cols' not in self.data.f_cols:
             self.data.input_shape = (self.data.seq_depth, self.data.n_x_cols)
 
+        self.data.targets_shape = (self.data.seq_depth, self.data.n_y_cols)
+        self.data.tracking_shape = (self.data.seq_depth, self.data.n_t_cols)
+
     def split_bars(self, _types):
 
         train_days = self.data.store['train']['config']['n_days']
@@ -642,7 +661,6 @@ class Trees():
             val_ed_day = val_st_day + val_days
 
         # --- reduce full bars by number of days --- #
-
         self.data.store['train']['full_bars'] = \
             self.data.full_bars[(self.data.full_bars['day_idx'] >= (train_st_day - self.data.seq_days)) &
                                 (self.data.full_bars['day_idx'] < train_ed_day)].copy()
@@ -693,7 +711,7 @@ class Trees():
         # -- normalise full bars -- #
         full_bars = self.normalize_bars(full_bars)
 
-        # === append seq bars to close bars to generate model bars === #
+        # --- append seq bars to close bars to generate model bars --- #
         for _idx in range(n_bars):
 
             _bar = close_bars.iloc[_idx]
@@ -706,11 +724,6 @@ class Trees():
 
             model_bars.append(seq_bars.to_numpy())
 
-            # self.data.setup_log['close_loop'].append(
-            #     {'specs': {'bar': _bar, 'st_idx':_st_idx ,'ed_idx':_ed_idx},
-            #      'seq_bars': seq_bars,
-            #      'model_bars': model_bars})
-
         # -- concat model bars -- #
         model_bars = np.concatenate(model_bars, axis=0)
 
@@ -754,6 +767,7 @@ class Trees():
 
             # -- x groups on -- #
             if self.network.x_groups_on:
+
                 # -- pr cols -- #
                 pr_cols = self.data.x_groups['pr_group']
 
@@ -799,6 +813,13 @@ class Trees():
 
         return data_bars
 
+    def setup_outputs_ds(self, data_type):
+
+        x_bars = self.data.store[data_type]['model_bars'][self.data.slice_configs['x_slice']]
+
+        return tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(x_bars,
+                                                                       dtype=self.data.dtype))
+
     def pre_build_datasets(self, data_type):
 
         batch_size = self.data.store[data_type]['config']['batch_size']
@@ -809,12 +830,12 @@ class Trees():
 
         # -- splts -- #
 
-        self.data.store[data_type]['x_bars'] = self.data.store[data_type]['model_bars'][..., :32]
-        self.data.store[data_type]['c_bars'] = self.data.store[data_type]['model_bars'][..., 32:39]
-        self.data.store[data_type]['y_bars'] = self.data.store[data_type]['model_bars'][..., 39:51]
-        self.data.store[data_type]['t_bars'] = self.data.store[data_type]['model_bars'][..., 51:]
+        self.data.store[data_type]['x_bars'] = self.data.store[data_type]['model_bars'][self.data.slice_configs['x_slice']]
+        self.data.store[data_type]['c_bars'] = self.data.store[data_type]['model_bars'][self.data.slice_configs['c_slice']]
+        self.data.store[data_type]['y_bars'] = self.data.store[data_type]['model_bars'][self.data.slice_configs['y_slice']]
+        self.data.store[data_type]['t_bars'] = self.data.store[data_type]['model_bars'][self.data.slice_configs['t_slice']]
 
-        if 'c_cols' in self.data.feature_cols:
+        if 'c_cols' in self.data.f_cols:
             self.data.store[data_type]['x_bars'] = \
                 np.concatenate([self.data.store[data_type]['x_bars'],
                                 self.data.store[data_type]['c_bars']], axis=-1)
@@ -827,8 +848,11 @@ class Trees():
 
         t_ds = tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(self.data.store[data_type]['t_bars'],
                                                                        dtype=self.data.dtype))
-
-        self.data.store[data_type]['data_ds'] = tf.data.Dataset.zip((x_ds, y_ds, t_ds)).batch(batch_size)
+        if self.data.outputs_on:
+            o_ds = self.setup_outputs_ds(data_type)
+            self.data.store[data_type]['data_ds'] = tf.data.Dataset.zip((x_ds, y_ds, t_ds, o_ds)).batch(batch_size)
+        else:
+            self.data.store[data_type]['data_ds'] = tf.data.Dataset.zip((x_ds, y_ds, t_ds)).batch(batch_size)
 
         # self.data.store[data_type]['data_ds'] = self.data.store[data_type]['data_ds'].prefetch(10)
 
@@ -969,6 +993,8 @@ class Chains():
 
         self.model_subs = []
 
+        self.logs = self.Logs()
+
     def build_chain(self):
 
         self.model_config = self.branch.configs.by_chain_idx('model', self.chain_idx)
@@ -1018,6 +1044,16 @@ class Chains():
                                              dtype=self.network.float_x,
                                              name=self.name + '_input_data')
 
+            self.targets_data = tf.keras.Input(shape=self.trees[0].data.targets_shape,
+                                               batch_size=self.trees[0].data.batch_size,
+                                               dtype=self.network.float_x,
+                                               name=self.name + '_targets_data')
+
+            self.tracking_data = tf.keras.Input(shape=self.trees[0].data.tracking_shape,
+                                                batch_size=self.trees[0].data.batch_size,
+                                                dtype=self.network.float_x,
+                                                name=self.name + '_tracking_data')
+
         # --- setup gauge/models --- #
         for model_idx in range(self.n_models):
 
@@ -1097,6 +1133,11 @@ class Chains():
         self.data_config['targets']['true_slice'] = np.s_[...,:split_on]
         self.data_config['targets']['tracking_slice'] = np.s_[..., split_on:]
 
+    @dataclass
+    class Logs:
+
+        batch_data: List = field(default_factory=lambda: {'main':[[]],'val':[[]]})
+
 class Gauge():
 
     def __init__(self, chain, model_idx):
@@ -1154,7 +1195,9 @@ class Gauge():
                          'val': None,
                          'test': None}
 
-        # -- setup strats -- #
+        self.log_theta = False
+        self.logs = self.Logs()
+
         self.setup_strats()
 
     def setup_strats(self):
@@ -1243,6 +1286,10 @@ class Gauge():
         # -- compile model -- #
         self.src.compile(optimizer=self.opt_fn)
 
+        self.log_theta = self.model_args['log_config']['log_theta']
+
+        self.theta = Theta(self)
+
         self.is_compiled = True
 
     def setup_cp(self, step_idx, load_cp):
@@ -1295,104 +1342,34 @@ class Gauge():
         if not self.is_compiled:
             self.compile_gauge()
 
-        self.src.pre_build(input_data=self.chain.input_data)
+        self.src.pre_build(input_data=self.chain.input_data,
+                           targets_data=self.chain.targets_data,
+                           tracking_data=self.chain.tracking_data)
 
-    def run_steps(self):
+    @dataclass
+    class Logs:
 
-        # self.run_train_steps(inputs)
-
-        return
-
-    def run_train_steps_(self, inputs, targets, batch_idx):
-
-        # tf.print('-- ',self.name, self.run_train_steps.experimental_get_tracing_count())
-                 # , batch_idx, inputs.shape, targets.shape, '\n')
-
-        batch_size = self.chain.branch.trees[0].data.batch_size
-
-        # --- load tape --- #
-        with tf.GradientTape() as tape:
-
-            # -- run model -- #
-            step_data = self.src(inputs=inputs, training=True, batch_idx=batch_idx)
-
-            # -- slice y_true -- #
-            step_data['y_true'] = targets[self.chain.data_config['targets']['true_slice']]
-
-            # -- slice y_tracking
-            step_data['y_tracking'] = targets[self.chain.data_config['targets']['tracking_slice']]
-
-            # -- step loss -- #
-            step_data['step_loss'] = self.loss_fn(step_data['y_true'], step_data['y_hat'])
-            step_data['step_loss'] = tf.nn.compute_average_loss(step_data['step_loss'], global_batch_size=batch_size)
-
-            # -- model loss -- #
-            step_data['model_loss'] = sum(self.src.losses)
-
-            # -- full loss -- #
-            step_data['full_loss'] = tf.add(step_data['step_loss'], step_data['model_loss'])
-
-            # -- run grads -- #
-            step_data['step_grads'] = tape.gradient(step_data['step_loss'], self.src.trainable_variables)
-
-            # -- apply grads -- #
-            self.src.optimizer.apply_gradients(zip(step_data['step_grads'], self.src.trainable_variables))
-
-            # -- save lr -- #
-            step_data['learning_rates'] = self.src.optimizer.lr_sch.cur_lr
-
-        return step_data
-
-    def run_preds_steps_(self):
-
-        tf.print('-- VAL --- ',self.name, self.run_preds_steps.python_function)
-                 # batch_idx, inputs.shape, targets.shape, '\n')
-
-        # batch_size = self.chain.branch.trees[0].data.batch_size
-
-        # -- run model -- #
-        # step_data = self.src(inputs=inputs, training=False, batch_idx=batch_idx)
-
-        some_data = ['sss']
-
-        # # -- slice y_true -- #
-        # step_data['y_true'] = targets[model.chain.data_config['targets']['true_slice']]
-        #
-        # # -- slice y_tracking
-        # step_data['y_tracking'] = targets[model.chain.data_config['targets']['tracking_slice']]
-        #
-        # # -- step loss -- #
-        # step_data['step_loss'] = model.gauge.loss_fn(step_data['y_true'], step_data['y_hat'])
-        # step_data['step_loss'] = tf.nn.compute_average_loss(step_data['step_loss'], global_batch_size=batch_size)
-        #
-        # # -- model loss -- #
-        # step_data['model_loss'] = sum(model.losses)
-        #
-        # # -- full loss -- #
-        # step_data['full_loss'] = tf.add(step_data['step_loss'], step_data['model_loss'])
-        #
-        # step_data['step_grads'] = None
-        # step_data['learning_rates'] = None
-
-        return some_data
+        calls: List = field(default_factory=lambda: {'main':[[]],'val':[[]]})
+        layers: List = field(default_factory=lambda: {'main':[[]],'val':[[]]})
+        run_data: List = field(default_factory=lambda: {'main':[[]],'val':[[]]})
+        theta: List = field(default_factory=lambda: [[]])
 
 class Theta:
 
-    def __init__(self, run):
+    def __init__(self, gauge):
 
-        self.run = run
+        self.gauge = gauge
+        self.logs_on = self.gauge.log_theta
 
         self.params = {'model_pre': [], 'model_post': [], 'opt': [], 'grads': []}
 
-        self.logs = []
-
     def save_params(self, param_type, grads=None):
 
-        if self.run.cur_chain.logs_on['theta']:
+        if self.logs_on:
 
             if param_type == 'model_pre' or param_type == 'model_post':
 
-                for idx, p in enumerate(self.run.model.trainable_variables):
+                for idx, p in enumerate(self.gauge.src.trainable_variables):
 
                     p_name = p.name
 
@@ -1417,7 +1394,7 @@ class Theta:
 
             if param_type == 'opt':
 
-                for idx, opt in enumerate(self.run.model.optimizer.get_weights()):
+                for idx, opt in enumerate(self.gauge.src.optimizer.get_weights()):
 
                     if idx > 0:
                         opt_data = {'idx': idx,
@@ -1427,15 +1404,13 @@ class Theta:
 
                         self.params['opt'].append(opt_data)
 
-    def log_params(self):
+    def log_params(self, epoch_idx):
 
-        if self.run.cur_chain.logs_on['theta']:
+        if self.logs_on:
 
-            epoch_idx = self.run.epoch_idx
+            if len(self.gauge.logs.theta) <= epoch_idx:
+                self.gauge.logs.theta.append([])
 
-            if len(self.logs) <= epoch_idx:
-                self.logs.append([])
-
-            self.logs[epoch_idx].append(self.params.copy())
+            self.gauge.logs.theta[epoch_idx].append(self.params.copy())
 
             self.params = {'model_pre': [], 'model_post': [], 'opt': [], 'grads': []}
\ No newline at end of file
