@@ -3,6 +3,8 @@ import tensorflow as tf
 import numpy as np
 import pandas as pd
 
+from photon.utils import list_idx_append
+
 from concurrent import futures
 from dataclasses import replace as dc_replace
 
@@ -123,13 +125,15 @@ class Gamma():
             inputs = data[0]
             targets = data[1]
             tracking = data[2]
+            outputs = data[3]
 
             if not chain.src.data_config['targets']['is_seq']:
                 targets = data[1][:, -1, :]
 
             batch_data = {'inputs': inputs,
                           'targets': targets,
-                          'tracking': tracking}
+                          'tracking': tracking,
+                          'outputs': outputs}
 
             yield batch_idx, batch_data
 
@@ -214,6 +218,7 @@ class Gamma():
 
             chain = self.update_live_chain(chain)
             model = self.update_live_model(model, chain=chain)
+            log_type = 'main'
 
         # -- flip -- #
         if is_val:
@@ -224,6 +229,7 @@ class Gamma():
                                            is_val=True)
 
             model = self.update_live_model(model, chain)
+            log_type = 'val'
 
         # -- loop batch data -- #
         for batch_idx, batch_data in self.get_data(chain, is_val):
@@ -231,6 +237,12 @@ class Gamma():
             chain.live.batch_idx = batch_idx
             model.live.batch_idx = batch_idx
 
+            # --- log batch data -- #
+            if chain.src.data_config['log_config']['log_batch_data'][log_type] and chain.chain_idx == 0:
+
+                list_idx_append(chain.src.logs.batch_data[log_type], chain.live.epoch_idx)
+                chain.src.logs.batch_data[log_type][chain.live.epoch_idx].append(batch_data)
+
             self.run_steps(model, batch_data, batch_idx, device_idx)
 
         # -- flip back -- #
@@ -242,6 +254,7 @@ class Gamma():
                                            is_val=False)
 
             model = self.update_live_model(model, chain)
+            log_type = 'main'
 
         return
 
@@ -259,7 +272,15 @@ class Gamma():
 
                 with tf.device('/GPU:' + str(device_idx)):
 
-                    step_data = model.src(inputs=batch_data['inputs'], training=True, batch_idx=batch_idx)
+                    step_data = model.src(inputs=batch_data['inputs'],
+                                          training=True,
+                                          batch_idx=batch_idx,
+                                          targets=batch_data['targets'],
+                                          tracking=batch_data['tracking'])
+
+                    # -- save pre model variables to theta -- #
+                    model.gauge.theta.save_params('model_pre')
+
                     step_data = self.run_splits(model, step_data, batch_data)
                     step_data = self.run_loss(model, step_data)
 
@@ -269,19 +290,31 @@ class Gamma():
                     # -- apply grads -- #
                     model.src.optimizer.apply_gradients(zip(step_data['step_grads'], model.src.trainable_variables))
 
+                    # -- save optimizer variables to theta -- #
+                    model.gauge.theta.save_params('opt')
+
+                    # -- save post model variables to theta -- #
+                    model.gauge.theta.save_params('model_post', step_data['step_grads'])
+
+                    # -- log theta params -- #
+                    model.gauge.theta.log_params(model.live.epoch_idx)
+
                     # -- save lr -- #
                     step_data['learning_rates'] = model.src.optimizer.lr_sch.cur_lr
 
-                    # model.chain.step_data_logs.append(step_data)
-
-                    # ---- non-training steps ---- #
+        # ---- non-training steps ---- #
         if model.live.run_type in ['infer', 'pred', 'val']:
 
             model.live.is_training = False
 
             with tf.device('/GPU:' + str(device_idx)):
 
-                step_data = model.src(inputs=batch_data['inputs'], training=True, batch_idx=batch_idx)
+                step_data = model.src(inputs=batch_data['inputs'],
+                                      training=True,
+                                      batch_idx=batch_idx,
+                                      targets=batch_data['targets'],
+                                      tracking=batch_data['tracking'])
+
                 step_data = self.run_splits(model, step_data, batch_data)
                 step_data = self.run_loss(model, step_data)
 
